#ifdef NSF
    #include "nsf.h"
#else
    #include "nes.h"
#endif

// Global
// ======

// 8-bit random number generator defined in coltrane.asm
// must initialize prng_seed to nonzero value
extern uint8 prng_seed;
#pragma zpsym("prng_seed");
uint8 __fastcall__ prng(void);

// Music stuff
// ===========
// ===========

#include "music.h"

// change this to start the music in a different section
#define START_SECTION 0

uint8 randomize;
uint8 random_count; // count of sections played (see RANDOM_CAP)

// maximum number of random tests to try for each bar
#define RANDOM_TESTS 4

// when to try to end the randomness (see random_count)
#define RANDOM_CAP  20

// customizable macros
// ===================

// vibrato

const sint8 flat[] = { 0, LOOP, 1 };
const sint8 vibrato[] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0,-1,-1, 0, 0, 0, 1, 1, 0, 0,
   -1,-2,-2,-1, 0, 1, 2, 2, 1, 0,
    LOOP, 10 };

// duty/volume

const uint8 rest_vm[]     = { 0x00, LOOP, 1 };
const uint8 sax_vm_f[]    = { 0x47, 0x0B, 0x0C, 0x0D, 0x0D, 0x0C, 0x0C, 0x0B, LOOP, 1 };
const uint8 sax_vm_p[]    = { 0x44, 0x07, 0x08, 0x08, 0x07, 0x07, 0x06, 0x05, LOOP, 1 };
const uint8 psolo_vm_f[]  = { 0xCF, 0xCD, 0xCB, 0xCA, 0xC9, 0xC8,
                              0xC7, 0xC7,
                              0xC6, 0xC6,
                              0xC5, 0xC5, 0xC5,
                              0xC4, 0xC4, 0xC4, 0xC4,
                              0xC3, 0xC3, 0XC3, 0xC3, 0xC3,
                              0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0xC2,
                              0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1,
                              0xC0, LOOP, 1 };
const uint8 psolo_vm_p[]  = { 0xC9, 0xC8, 0xC7,
                              0xC6, 0xC6,
                              0xC5, 0xC5, 0xC5,
                              0xC4, 0xC4, 0xC4,
                              0xC3, 0xC3, 0xC3, 0xC3, 0xC3,
                              0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0XC2, 0xC2, 0xC2, 0xC2,
                              0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1,
                              0xC0, LOOP, 1 };
const uint8 pchord_vm_f[] = { 0xCF, 0xCE, 0xCD, 0xCC, 0xCB, 0xCA, 0xC9,
                              0xC8, 0xC8,
                              0xC7, 0xC7,
                              0xC6, 0xC6, 0xC6,
                              0xC5, 0xC5, 0xC5,
                              0xC4, 0xC4, 0xC4, 0xC4,
                              0xC3, 0xC3, 0XC3, 0xC3, 0xC3,
                              0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0xC2,
                              0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1,
                              0xC0, LOOP, 1 };
const uint8 pchord_vm_p[] = { 0xC9, 0xC8,
                              0xC7, 0xC7,
                              0xC6, 0xC6,
                              0xC5, 0xC5, 0xC5,
                              0xC4, 0xC4, 0xC4,
                              0xC3, 0xC3, 0xC3, 0xC3, 0xC3,
                              0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0xC2, 0XC2, 0xC2, 0xC2, 0xC2,
                              0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1,
                              0xC0, LOOP, 1 };

// drums (volume, period)

const uint8 ride_vm[]  = { 10, 11, 10,  9, 8, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 2, 1, LOOP, 1 };
const uint8 ride_pm[]  = { 0x80, 0x01, LOOP, 1 };

const uint8 hat_vm[]   = { 13, 8, 5, 2, 1, LOOP, 1 };
const uint8 hat_pm[]   = { 0x00, 0x01, LOOP, 1 };

// alternate hat (light snareish) randomly replaces hat
const uint8 ahat_vm[]  = { 6, 8, 5, 3, 2, 1, LOOP, 1 };
const uint8 ahat_pm[]  = { 0x05, 0x01, 0x02, 0x03, 0x04, LOOP, 1 };

const uint8 snare_vm[] = { 12, 10, 9, 8, 7, 6, 5, 4, 3, 2, 2, 1, LOOP, 1 };
const uint8 snare_pm[] = { 0x06, 0x03, 0x01, 0x00, 0x01, 0x02, 0x03, 0x04, LOOP, 1 };

const uint8 hride_vm[] = { 10, 11, 10,  9, 8, 7, 6, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, LOOP, 1 };
const uint8 hride_pm[] = { 0x80, 0x01, 0x01, 0x02, LOOP, 1 };

// first note of tail is 51 frames long (ride decay)
// followed by drum roll in 3s, loops quietly to end
const uint8 tail_vm[]  = { 10, 13, 14, 13, 12, 11, 11, 10, 10, 9,
                           9, 9, 8, 8, 8, 7, 7, 7, 6, 6,
                           6, 5, 5, 5, 5, 4, 4, 4, 4, 4,
                           3, 3, 3, 3, 3, 3, 2, 2, 2, 2,
                           2, 2, 2, 2, 2, 2, 2, 1, 1, 1,
                           1,
                           7, 5, 4, // roll begins
                           7, 5, 4,
                           6, 5, 4,
                           5, 4, 3,
                           4, 3, 2,
                           4, 3, 2,
                           3, 2, 1, // loop
                           3, 2, 1,
                           3, 2, 1,
                           3, 2, 1,
                           2, 2, 1,
                           2, 2, 1,
                           2, 2, 1,
                           LOOP, 21 };
const uint8 tail_pm[]  = { 0x80, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1,
                           6, 3, 1, // roll loop
                           LOOP, 3 };

// ending snare, terminates on 0
const uint8 end_vm[]   = { 7, 5, 4, 3, 3, 2, 2, 1, 1, 0, LOOP, 1 };
const uint8 end_pm[]   = { 0x06, 0x03, 0x02, 0x03, 0x04, LOOP, 1 };

// music sequences not in music.h
// ==============================

#define DR  12 // ride
#define DH  13 // hat
#define DS  14 // snare
#define DC  15 // head ride (heavier)
#define DT  16 // tail roll

// standard drum pattern
const uint8 drum0[] = {DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,BAR};

const uint8 drumH0[] = { DC,24,DC,24,DC,24,DC,18,DC, 6,BAR};
const uint8 drumH1[] = {TIE,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DC,18,DC,30,BAR};
const uint8 drumH2[] = { DC,24,DC,24,DC,24,DC,18,DC, 6,BAR};
const uint8 drumH3[] = {TIE,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,BAR};
const uint8 drumH4[] = { DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,BAR};
const uint8 drumH5[] = { DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,BAR};
const uint8 drumH6[] = { DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,BAR};
const uint8 drumH7[] = { DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,DR,18,DS,30,BAR};
const uint8 drumH8[] = { DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,DR,12,DH, 8,DH, 4,DS,12,DH, 8,DH, 4,BAR}; // second head continues

const uint8 drumT6[] = { DR,12,DH, 8,DH, 4,DC,12,DC,12,DC,24,DC,18,DT, 6,BAR};
const uint8 drumT7[] = {TIE,250,BAR}; // ends before, just rolls out on DT macro

const uint8* drumHead0[8] = { drumH0, drumH1, drumH2, drumH3, drumH4, drumH5, drumH6, drumH7 };
const uint8* drumHead1[8] = { drumH0, drumH1, drumH2, drumH3, drumH4, drumH5, drumH6, drumH8 };
const uint8* drumTail[8]  = { drumH0, drumH1, drumH2, drumH3, drumH4, drumH5, drumT6, drumT7 };

// piano head
const uint8 pcHead0[] = { 12,24, 13,24, 14,24, 15,18, 16, 6,BAR};
const uint8 pcHead1[] = {TIE,36,RES,12, 17,18, 18,18,RES,12,BAR};
const uint8 pcHead2[] = { 19,24, 20,24, 21,24, 22,18, 23, 6,BAR};
const uint8 pcHead3[] = {TIE,36,RES,12, 24,24, 25,18, 26, 6,BAR};
const uint8 pcHead4[] = {TIE,36,RES,12, 27,24, 28,18, 29, 6,BAR};
const uint8 pcHead5[] = {TIE,36,RES,12, 30,24, 31,18, 32, 6,BAR};
const uint8 pcHead6[] = {TIE,36,RES,12, 33,24, 34,18, 35, 6,BAR};
const uint8 pcHead7[] = {TIE,36,RES,12, 36,18, 37, 6,RES,24,BAR};
const uint8 pcTail7[] = {TIE,250,BAR}; // note: ends before sax

const uint8* pcHead[8] = { pcHead0, pcHead1, pcHead2, pcHead3, pcHead4, pcHead5, pcHead6, pcHead7 };
const uint8* pcTail[8] = { pcHead0, pcHead1, pcHead2, pcHead3, pcHead4, pcHead5, pcHead6, pcTail7 };

// global static music variables
// =============================

static uint8  note;
static uint16 freq;

#define SECTIONS 21
#define BARS 8
#define BEATS (4*24)

uint8 section;
uint8 bar;
uint8 beat; // frame of bar

// nesmain NMI tick (run at 60hz)
// ==============================

uint8 apu_cur[16];  // current state of APU $4000-400F
uint8 apu_buf[16];  // next APU state to send
uint8 apu_send = 0; // don't send until this != 0

void nmi_sound()
{
    // if apu_send is true, copy apu_buf into apu_cur, writing changes to APU registers
    
    if (!apu_send)
        return;

    for (i=0; i<16; ++i)
    {
        if (apu_cur[i] != apu_buf[i])
        {
            apu_cur[i] = apu_buf[i];
            RAW_BUS(0x4000+i) = apu_cur[i];
        }
    }
    RAW_BUS(0x4017) = 0xC0; // synch the frame counter
    apu_send = 0;
}

// drums (noise)
// =============

// state variables
uint8 drum_beat;
uint8 drum_length;
const uint8* drum_macro_vol; // volume macro
const uint8* drum_macro_per; // period macro
const uint8* drum_data;

// play variables
uint8 drum_volume;
uint8 drum_period;

void set_data_drum()
{
    drum_beat = 0;
    switch(section)
    {
        case (0):
            drum_data = drumHead0[bar];
            break;
        case (1):
            drum_data = drumHead1[bar];
            break;
        case (SECTIONS-2):
            drum_data = drumHead0[bar];
            break;
        case (SECTIONS-1):
            drum_data = drumTail[bar];
            break;
        default:
            drum_data = drum0; // standard drum pattern
    }
}

void setup_drum()
{
    drum_beat = 0;
    drum_length = 0;
    
    drum_macro_vol = ride_vm;
    drum_macro_per = ride_pm;
    
    drum_volume = 0;
    drum_period = 0;
    
    set_data_drum();
}

void play_drum()
{
    apu_buf[0xC] = drum_volume | 0x30;
    apu_buf[0xE] = drum_period;
    apu_buf[0xF] = 0x08;
}

void tick_drum_macro()
{
    drum_volume = *drum_macro_vol;
    if (drum_volume == LOOP)
    {
        drum_macro_vol -= drum_macro_vol[1];
        drum_volume = *drum_macro_vol;
    }
    else
        ++drum_macro_vol;
    
    drum_period = *drum_macro_per;
    if (drum_period == LOOP)
    {
        drum_macro_per -= drum_macro_per[1];
        drum_period = *drum_macro_per;
    }
    else
        ++drum_macro_per;
}

void tick_drum()
{
    // still playing a note, do nothing
    if (drum_length > 0)
    {
        --drum_length;
        tick_drum_macro();
        return;
    }
    
  //read_note:
    note = drum_data[drum_beat];
    ++drum_beat;
    
    if (note == BAR)
    {
        --drum_beat; // rewind and hang on BAR
        tick_drum_macro();
        return;
    }
    
    drum_length = drum_data[drum_beat] - 1;
    ++drum_beat;
    switch (note)
    {
        case TIE:
            break;
        case RES:
            drum_macro_vol = rest_vm;
            break;
        case DR:
            drum_macro_vol = ride_vm;
            drum_macro_per = ride_pm;
            break;
        case DH:
            if ((prng() & 15) >= 11)
            {
                drum_macro_vol = ahat_vm;
                drum_macro_per = ahat_pm;
            }
            else
            {
                drum_macro_vol = hat_vm;
                drum_macro_per = hat_pm;
            }
            break;
        case DS:
            drum_macro_vol = snare_vm;
            drum_macro_per = snare_pm;
            break;
        case DC:
            drum_macro_vol = hride_vm;
            drum_macro_per = hride_pm;
            break;
        case DT:
            drum_macro_vol = tail_vm;
            drum_macro_per = tail_pm;
            break;
        default: // unknown
            drum_macro_vol = rest_vm;
            break;
    }
    tick_drum_macro();
}

// piano chords (square 0)
// =======================

// state variables
uint8 pchord_beat;
uint8 pchord_length;
uint8 pchord_ghost;
const sint8* pchord_macro_arp; // arp macro
const uint8* pchord_macro_vol; // volume/duty macro
const uint8 * pchord_data;

// play variables
uint8 pchord_note;
uint8 pchord_volume; // combined volumne/duty register, 0 if note off

void set_data_pchord()
{
    pchord_beat = 0;
    if (section < 2 || section == (SECTIONS-2))
        pchord_data = pcHead[bar];
    else if (section == (SECTIONS-1))
        pchord_data = pcTail[bar];
    else if (section >= 13 && section < 17)
        pchord_data = pianoB[section][bar];
    else
    {
        pchord_data = pianoBA[bar];
    }
}

void setup_pchord()
{
    pchord_beat = 0;
    pchord_length = 0;
    
    pchord_ghost = 0;
    
    pchord_macro_arp = slide[0];
    pchord_macro_vol = rest_vm;
    
    pchord_note = 0;
    pchord_volume = 0;
    
    set_data_pchord();
}

void play_pchord()
{
    freq = pitch[pchord_note];
    apu_buf[0x0] = pchord_volume | 0x30; // disable length counter / envelope
    apu_buf[0x2] = (uint8)(freq);
    apu_buf[0x3] = (uint8)(freq >> 8) | 0x80;
}

void tick_pchord_macro()
{
    pchord_volume = *pchord_macro_vol;
    if (pchord_volume == LOOP)
    {
        pchord_macro_vol -= pchord_macro_vol[1];
        pchord_volume = *pchord_macro_vol;
    }
    else
        ++pchord_macro_vol;
    
    pchord_note = *pchord_macro_arp;
    if (pchord_note == LOOP)
    {
        pchord_macro_arp -= pchord_macro_arp[1];
        pchord_note = *pchord_macro_arp;
    }
    else
        ++pchord_macro_arp;
}

void tick_pchord()
{
    // still playing a note, do nothing
    if (pchord_length > 0)
    {
        --pchord_length;
        tick_pchord_macro();
        return;
    }
    
  read_note:
    note = pchord_data[pchord_beat];
    ++pchord_beat;
    
    if (note == BAR)
    {
        --pchord_beat; // rewind and hang on BAR
        tick_pchord_macro();
        return;
    }
    else if(note == GHO)
    {
        pchord_ghost = 1;
        goto read_note;
    }
    
    pchord_length = pchord_data[pchord_beat] - 1;
    ++pchord_beat;
    switch (note)
    {
        case TIE:
            break;
        case RES:
            pchord_macro_vol = rest_vm;
            break;
        default:
            pchord_macro_arp = chord[note];
            pchord_macro_vol = pchord_ghost ? pchord_vm_p : pchord_vm_f;
            pchord_ghost = 0;
            break;
    }
    tick_pchord_macro();
}

// bass (triangle)
// ===============

// state variables
uint8 bass_beat;
uint8 bass_length;
const uint8* bass_data;

// play variables
uint8 bass_note;
uint8 bass_on;

void set_data_bass()
{
    bass_beat = 0;
    bass_data = bass[section][bar];
}

void setup_bass()
{
    bass_beat = 0;
    bass_length = 0;
    
    bass_note = 0;
    bass_on = 0;

    set_data_bass();
}

void play_bass()
{
    freq = pitch[bass_note];

    apu_buf[0x8] = bass_on ? 0x81 : 0x00;
    apu_buf[0xA] = (uint8)(freq);
    apu_buf[0xB] = (uint8)(freq >> 8);
    
    // write these registers every frame
    apu_cur[0x8] = 0xFF;
    apu_cur[0xB] = 0xFF;
}

void tick_bass()
{
    // still playing a note, do nothing
    if (bass_length > 0)
    {
        --bass_length;
        return;
    }
  //read_note:
    note = bass_data[bass_beat];
    ++bass_beat;
    
    if (note == BAR)
    {
        --bass_beat; // rewind and hang on BAR
        //tick_bass_macro();
        return;
    }
    
    bass_length = bass_data[bass_beat] - 1;
    ++bass_beat;
    switch (note)
    {
        case TIE:
            break;
        case RES:
            bass_on = 0;
            break;
        default:
            bass_note = note;
            bass_on = 1;
            break;
    }
}

// sax + piano solo (square 1)
// ===========================

// state variables
uint8 sax_beat;
uint8 sax_length;
uint8 sax_psolo; // piano solo override
uint8 sax_gliss;  // where to start a slide
uint8 sax_ghost;
uint8 sax_note_ref;
sint8 sax_arp;
const sint8* sax_macro_arp; // arp macro
const sint8* sax_macro_ben; // bend macro
const uint8* sax_macro_vol; // volume/duty macro
const uint8* sax_data;

// play variables
uint8 sax_note;
sint8 sax_bend; // pitch bend
uint8 sax_volume; // combined volumne/duty register, 0 if note off
uint8 sax_piano; // indicates piano instrument in use

void set_data_sax()
{
    sax_beat = 0;
    
    if (section == 13 && bar == 1)
        sax_psolo = 1;
    if (section == 17 && bar == 0)
        sax_psolo = 0;
        
    if (!sax_psolo)
        sax_data = sax[section][bar];
    else
        sax_data = pianoT[section][bar];
}

void setup_sax()
{
    sax_beat = 0;
    sax_length = 0;
    sax_psolo = 0;
    
    sax_gliss = 0;
    sax_ghost = 0;
    
    sax_note_ref = 0;
    sax_arp = 0;
    sax_macro_arp = slide[0];
    sax_macro_ben = vibrato;
    sax_macro_vol = rest_vm;
    
    sax_note = 0;
    sax_bend = 0;
    sax_volume = 0;
    sax_piano = 0;
    
    set_data_sax();
}

void play_sax()
{
    freq = pitch[sax_note] + sax_bend;

    apu_buf[0x4] = sax_volume | 0x30; // disable length counter / envelope
    apu_buf[0x6] = (uint8)(freq);
    apu_buf[0x7] = (uint8)(freq >> 8) | 0x80;
}

void tick_sax_macro()
{
    sax_volume = *sax_macro_vol;
    if (sax_volume == LOOP)
    {
        sax_macro_vol -= sax_macro_vol[1];
        sax_volume = *sax_macro_vol;
    }
    else
        ++sax_macro_vol;
    
    sax_arp = *sax_macro_arp;
    if (sax_arp == LOOP)
    {
        sax_macro_arp -= sax_macro_arp[1];
        sax_arp = *sax_macro_arp;
    }
    else
        ++sax_macro_arp;
    sax_note = sax_note_ref + sax_arp;
    
    sax_bend = *sax_macro_ben;
    if (sax_bend == LOOP)
    {
        sax_macro_ben -= sax_macro_ben[1];
        sax_bend = *sax_macro_ben;
    }
    else
       ++sax_macro_ben;
}

void tick_sax()
{
    sax_piano = sax_psolo;

    // still playing a note, do nothing
    if (sax_length > 0)
    {
        --sax_length;
        tick_sax_macro();
        return;
    }
    
  read_note:
    note = sax_data[sax_beat];
    ++sax_beat;
    
    if (note == BAR)
    {
        --sax_beat; // rewind and hang on BAR
        tick_sax_macro();
        return;
    }
    else if(note == GHO)
    {
        sax_ghost = 1;
        goto read_note;
    }
    else if(note == SLI)
    {
        sax_gliss = sax_data[sax_beat];
        ++sax_beat;
        goto read_note;
    }
    
    sax_length = sax_data[sax_beat] - 1;
    ++sax_beat;
    switch (note)
    {
        case TIE:
            break;
        case RES:
            // piano solo notes are always let sustain (no need for rest)
            if (!sax_psolo)
                sax_macro_vol = rest_vm;
            break;
        default:
            sax_note_ref = note;
            if (!sax_psolo)
            {
                sax_macro_arp = sax_gliss ? slide[note-sax_gliss] : slide[0];
                sax_macro_ben = vibrato;
                sax_macro_vol = sax_ghost ? sax_vm_p : sax_vm_f;
            }
            else
            {
                sax_macro_arp = slide[0];
                sax_macro_ben = flat;
                sax_macro_vol = sax_ghost ? psolo_vm_p : psolo_vm_f;
            }
            sax_gliss = 0;
            sax_ghost = 0;
            break;
    }
    tick_sax_macro();
}

// Music main
// ==========

uint8 ring_buffer_pos; // read position for buffered playback ring buffer, see below
uint8 ring_buffer_next;

void setup_music()
{
    ring_buffer_pos = 0;
    ring_buffer_next = 1;

    // initialize APU
    for (i=0; i<16; ++i)
    {
        apu_cur[i] = 0xFF; // put
    }
    apu_buf[0x0] = 0x30; // S0 constant volume 0, length counter off
    apu_buf[0x1] = 0x00; // S0 sweep off
    apu_buf[0x2] = 0x00; // S0 freq low
    apu_buf[0x3] = 0x00; // S0 freq high
    apu_buf[0x4] = 0x30; // S1 constant volume 0, length counter off
    apu_buf[0x5] = 0x00; // S1 sweep off
    apu_buf[0x6] = 0x00; // S1 freq low
    apu_buf[0x7] = 0x00; // S1 freq high
    apu_buf[0x8] = 0x00; // T silence
    apu_buf[0x9] = 0x00; // T (unused)
    apu_buf[0xA] = 0x00; // T freq low
    apu_buf[0xB] = 0x00; // T freq high
    apu_buf[0xC] = 0x30; // N constant volume 0, length counter off
    apu_buf[0xD] = 0x00; // N (unused)
    apu_buf[0xE] = 0x00; // N loop off, period 0
    apu_buf[0xF] = 0x00; // N unused length counter
    apu_send = 1;
    nmi_sound(); // use nmi_sound to initialize APU
    RAW_BUS(0x4015) = 0x0F; // enable S0/S1/N/T
    RAW_BUS(0x4010) = 0x00; // DPCM IRQ disable
    RAW_BUS(0x4011) = 0x00; // DPCM counter to 0
    
    section = START_SECTION;
    bar = 0;
    beat = 0;

    setup_drum();
    setup_pchord();
    setup_bass();
    setup_sax();
}

void tick_all()
{
    tick_drum();
    tick_pchord();
    tick_bass();
    tick_sax();
    
    if (section < SECTIONS)
    {
        ++beat;
        if (beat >= BEATS)
        {
            beat = 0;
            ++bar;
            if (bar >= BARS)
            {
                bar = 0;
                ++section;
                ++random_count;
            }
            if (section < SECTIONS)
            {
                // randomized solo: selects a new sax section if the next note matches
                if (randomize && section >= 2 && section < 13)
                {
                    b = sax[section][bar][0]; // starting note
                    if (b != TIE) // section starting with TIE cannot jump
                    {
                        for (i=0; i < RANDOM_TESTS; ++i)
                        {
                            a = (prng() % 10) + 2; // a in range [2,12]
                            if (random_count > RANDOM_CAP && i == 0) // try to force the piano solo when it gets too long
                            {
                                a = (bar == 0) ? 13 : 12;
                            }
                            
                            if (a == section) continue; // that would be boring
                            if (sax[a][bar][0] == b) // found a matching note
                            {
                                // possibilities:
                                //   matching note (good)
                                //   matching rest (okay)
                                //   matching SLI (okay because all glisses at bar start are same)
                                //   matching GHO (doesn't occur)
                                section = a;
                                break;
                            }
                        }
                    }
                }
            
                set_data_drum();
                set_data_pchord();
                set_data_bass();
                set_data_sax();
            }
        }
    }
    if (section >= SECTIONS && section != 0xFF)
    {
        if (sax_length == 0 && sax_data[sax_beat] == BAR) // sax is finished final cadenza
        {
            // turn off anything still going
            drum_macro_vol = end_vm;
            drum_macro_per = end_pm;
            pchord_macro_vol = rest_vm;
            pchord_volume = 0;
            bass_on = 0;
            sax_macro_vol = rest_vm;
            sax_volume = 0;
            section = 0xFF; // end tune
        }
    }
}

void play_all()
{
    play_drum();
    play_pchord();
    play_bass();
    play_sax();
    apu_send = 1;
}

// for direct playback (use for NSF)

void tick_music()
{
    tick_all();
    play_all();
}

#ifndef NSF

// for buffered playback

#define RB_TILES   18
#define RB_LEN     (RB_TILES * 4)

uint8* const ring_buffer_var[10] =
{
    &drum_volume,      // 0
    &drum_period,      // 1
    &pchord_note,      // 2
    &pchord_volume,    // 3
    &bass_note,        // 4
    &bass_on,          // 5
    &sax_note,         // 6
    (uint8*)&sax_bend, // 7
    &sax_volume,       // 8
    &sax_piano         // 9
};

uint8 ring_buffer[10][RB_TILES * 4]; // store 4 notes per tile
void draw_music(); // forward declaration

#define RING_BUFFER_SWAP(x) \
    ring_buffer[x][ring_buffer_pos] = *ring_buffer_var[x]; \
    *ring_buffer_var[x] = ring_buffer[x][ring_buffer_next];

#define RING_BUFFER_RESTORE(x) \
    *ring_buffer_var[x] = ring_buffer[x][ring_buffer_pos];

void buffer_music()
{
    tick_all();
    
    // buffer and store the playback variables, then replace before calling play
    // NOTE: unrolling loop was important for performance reasons;
    ring_buffer_next = (ring_buffer_pos + 1) % RB_LEN;
    RING_BUFFER_SWAP(0);
    RING_BUFFER_SWAP(1);
    RING_BUFFER_SWAP(2);
    RING_BUFFER_SWAP(3);
    RING_BUFFER_SWAP(4);
    RING_BUFFER_SWAP(5);
    RING_BUFFER_SWAP(6);
    RING_BUFFER_SWAP(7);
    RING_BUFFER_SWAP(8);
    RING_BUFFER_SWAP(9);
    
    draw_music();
    
    play_all();
    
    // restore playback variables
    RING_BUFFER_RESTORE(0);
    RING_BUFFER_RESTORE(1);
    RING_BUFFER_RESTORE(2);
    RING_BUFFER_RESTORE(3);
    RING_BUFFER_RESTORE(4);
    RING_BUFFER_RESTORE(5);
    RING_BUFFER_RESTORE(6);
    RING_BUFFER_RESTORE(7);
    RING_BUFFER_RESTORE(8);
    RING_BUFFER_RESTORE(9);
    ring_buffer_pos = ring_buffer_next;
}

// Graphics stuff
// ==============

const uint8 palette[16] = {
    0x0F, 0x19, 0x27, 0x09,
    0x0F, 0x2D, 0x19, 0x30,
    0x0F, 0x29, 0x16, 0x11,
    0x0F, 0x27, 0x19, 0x29,
};

const uint8 piano_roll[6] = { 0x80, 0x90, 0x90, 0xA0, 0xA0, 0xA0 };

const uint8 piano_keys[4][12] = {{0xC8,0xC7,0xC6,0xC8,0xC6,0xC8,0xC7,0xC6,0xC8,0xC6,0xC8,0xC6},
                                 {0xC5,0xC4,0xC3,0xC5,0xC3,0xC5,0xC4,0xC3,0xC5,0xC3,0xC5,0xC3},
                                 {0xC0,0xC0,0xC2,0xC0,0xC2,0xC0,0xC0,0xC2,0xC0,0xC2,0xC0,0xC2},
                                 {0xC0,0xC0,0xC1,0xC0,0xC1,0xC0,0xC0,0xC1,0xC0,0xC1,0xC0,0xC1}};

const uint8* title = "GIANT STEPS";

const uint8 attribute_line[8] = { 0xFF, 0x0F, 0x00, 0x00, 0x00, 0x50, 0x55, 0x55 };

static uint8 keyboard_sprite; // index for cycling 8 keyboard sprites (5+ note chords will flicker)
static uint8 roll_note[2];
static uint8 roll_pos; // current roll column to write

uint16 nametable_offset(uint8 x, uint8 y)
{
    return x + (((uint16)y) << 5);
}

uint16 attribute_offset(uint8 x, uint8 y)
{
    return 0x3C0 + (x >> 2) + ((y >> 2) << 3);
}

void sprite(uint8 index, uint8 x, uint8 y, uint8 tile, uint8 pal)
{
    index <<= 2;
    ppu_oam[index+0] = y-1;
    ppu_oam[index+1] = tile;
    ppu_oam[index+2] = pal; // note: flipping and behind-background not implemented
    ppu_oam[index+3] = x;
}

void fill_screen()
{
    for (i=0;i<6;++i) // 6 lines of blank
        for (j=0;j<32;++j)
            ppu_write(0x20);
    for (i=0;i<16;++i) // 16 lines of piano roll
    {
        a = piano_roll[i%6];
        for (j=0;j<32;++j)
            ppu_write(a);
    }
    for (i=0;i<8;++i) // 8 lines of piano keys
    {
        a = (i<3) ? i : 3;
        for (j=0;j<32;++j)
            ppu_write(piano_keys[a][j%12]);
    }
    
    // attributes
    for (i=0;i<8;++i)
        for (j=0;j<8;++j)
            ppu_write(attribute_line[i]);
}

void setup_background()
{
    roll_pos = 16 + RB_TILES;
    
    // set palette
    for (i=0; i<16; ++i)
    {
        ppu_pal[i] = ppu_pal[i+16] = palette[i];
    }
    
    ppu_latch_h(0x2000);
    fill_screen();
    ppu_latch_h(0x2400);
    fill_screen();
    
    // title text
    ppu_latch_h(0x2000 + nametable_offset(12,3));
    while (*title)
    {
        ppu_write(*title);
        ++title;
    }
    
    sprite(0, 16, 176, 0xD8, 1); // sprite 0 hit (just below piano roll, same colour as background)
    sprite(1, 124, 39, 0xD3, 2); // arrow indicating now position
}

void no_roll()
{
    ppu_nam[0] = 0; // no nametable update this frame
}

void draw_roll()
{
    uint16 column = ((roll_pos >= 32) ? 0x2400 : 0x2000) + (roll_pos & 31);
    ppu_nam[0] = (6+16) | 0x80; // vertical write 6+16 bytes
    ppu_nam[3+6+16] = 0; // end of nametable updates
    
    ppu_nam[1] = (column >> 8);
    ppu_nam[2] = (column & 0xFF);
    
    // blank area, should wipe title
    for (i=0; i<6; ++i)
        ppu_nam[3+i] = 0x20;
    // blank keyboard area
    a = 0;
    for (i=0; i<16; ++i) // this is messing things up
    {
        ppu_nam[3+6+i] = piano_roll[a];
        ++a;
        if (a >= 6) a = 0;
    }
    
    // keyboard bitmask:
    //    lower left  = 0x01
    //    upper left  = 0x04
    //    lower right = 0x02
    //    upper right = 0x08
    
    if (ring_buffer[9][ring_buffer_pos]) // sax_piano
    {
        // drop the piano solo an octave on the display.
        roll_note[0] -= 12;
        roll_note[1] -= 12;
    }
    
    // set notes on the keyboard
    if (roll_note[0] >= 22 && roll_note[0] < 54)
    {
        a = 53 - roll_note[0];
        if (a & 1) // lower left
            ppu_nam[3+6+(a>>1)] |= 0x01;
        else       // upper left
            ppu_nam[3+6+(a>>1)] |= 0x04;
    }
    if (roll_note[1] >= 22 && roll_note[1] < 54)
    {
        a = 53 - roll_note[1];
        if (a & 1) // lower right
            ppu_nam[3+6+(a>>1)] |= 0x02;
        else       // upper right
            ppu_nam[3+6+(a>>1)] |= 0x08;
    }

    // increment roll column
    roll_pos = (roll_pos + 1) & 63;
}

void draw_music()
{
    // piano roll
    a = ring_buffer[6][ring_buffer_pos]; // sax_note
    b = ring_buffer[8][ring_buffer_pos]; // sax_volume
    switch (ring_buffer_pos & 3)
    {
        case 0:
            if (b) // if sax is on
                roll_note[0] = a; // read sax note
            else
                roll_note[0] = 0;
            no_roll();
            break;
        case 1:
            if (b) // replace note if changes (2 frames are combined into 1 half-tile)
                roll_note[0] = a;
            no_roll();
            break;
        case 2:
            if (b)
                roll_note[1] = a;
            else
                roll_note[1] = 0;
            no_roll();
            break;
        case 3:
            if (b)
                roll_note[1] = a;
            draw_roll();
            break;
    };

    // bassline
    sprite(2, 8*(bass_note-11), bass_on ? 195 : 255, 0xD1, 2);
    
    // keyboard
    keyboard_sprite = (keyboard_sprite + 1) & 7;
    a = pchord_note;
    if (a < 23) // wrap low notes into lowest shown octave
        a += 12;
    else if (a > 54) // wrap high notes into highest shown octave
        a -= 12;
    a -= 23;
    if (pchord_volume)
        sprite(3 + keyboard_sprite, 8*a, 204, 0xD2, 2); // draw another chord note
    else
        for (i=0;i<8;++i)
            sprite(3 + i, 255, 255, 0xD2, 2); // hide whole chord
}

// Title Screen
// ============

void title_text(uint8 x, uint8 y, const char* text)
{
    ppu_latch_h(0x2000 + nametable_offset(x,y));
    while (*text)
    {
        ppu_write(*text);
        ++text;
    }
}

void title_text_centered(uint8 y, const char* text)
{
    const char* t = text;
    a = 0;
    while (*t)
    {
        ++a;
        ++t;
    }
    title_text(16 - (a>>1), y, text);
}

void title_screen()
{
    // black and white palette
    for (i=0; i < 32; ++i)
        ppu_pal[i] = ((i & 3) == 3) ? 0x30 : 0x0F;
    // cls
    ppu_latch_h(0x2000);
    for (k=0;k<0x400;++k)
        ppu_write(0x00);
    // text
    title_text_centered( 8,"GIANT STEPS");
    title_text_centered(14,"START");
    title_text_centered(16,"RANDOMIZE");
    title_text_centered(19,"BRADSMITH, 2012");
    title_text_centered(21,"ORIGINAL BY           ");
    title_text_centered(22,"   JOHN COLTRANE, 1960");
    title_text_centered(26,"HTTP://RAINWARRIOR.CA");
    // hide the randomize text
    ppu_pal[7] = 0x0F;
    ppu_latch_h(0x2000 + attribute_offset(0,16));
    for (i=0; i < 8; ++i)
        ppu_write(0x05);
    // sprite next to START
    ppu_pal[19] = 0x29;
    sprite(0,(12*8)+4,14*8,0xD2,0x00);

    ppu_send = 1;
    ppu_nmi_on();
    
    // wait for any held button to be released before allowing input
    gamepad_poll();
    while(gamepad0 != 0)
        gamepad_poll();
    
    // wait for input
    while(1)
    {
        gamepad_poll();
        
        if (gamepad0 & PAD_SELECT)
        {
            // rainbow flash the "RANDOMIZE" text
            b = (b + 1) & 0xF;
            ppu_pal[7] = b | 0x20;
            // tick random number generator
            prng();
        }
        else
        {
            // hide the randomize text
            ppu_pal[7] = 0x0F;
        }
        
        if (gamepad0 & (PAD_START | PAD_A | PAD_B))
            break;

        ppu_send = 1;
        ppu_nmi_wait();
    }

    if (0 == (gamepad0 & PAD_SELECT))
    {
        // don't randomize
        prng_seed = 1;
        randomize = 0;
    }
    else
    {
        // randomize
        randomize = 1;
        random_count = 0;
    }
    
    ppu_nmi_wait();
    ppu_nmi_off();
    ppu_render_off();

    // wait for start to be released
    while (gamepad0 != 0)
        gamepad_poll();
}

// NES Main
// ========

// the following functions are called directly from nesmain

static uint8 scroll;

void main()
{
    prng_seed = 1;
    
    title_screen();

    setup_music();
    setup_background();
    
    ppu_send = 1;
    ppu_nmi_on();
    
    while (!(gamepad0 & PAD_SELECT))
    {
        buffer_music();
        gamepad_poll();
        //debug_grey(); // for timing the code
        
        ++scroll;
        ppu_ctrl = (ppu_ctrl & ~0x01) | ((scroll & 0x80) ? 0x01 : 0x00);
        ppu_scroll_x = scroll<<1;
        ppu_scroll_y = 0;
        ppu_send = 1;
        
        // scroll split for keyboard (note: same keyboard is on both nametable pages,
        // so ppu_ctrl bit does not matter)
        ppu_sprite_0_hit();
        RAW_BUS(0x2005) = 0;
        RAW_BUS(0x2005) = 0;
        
        ppu_nmi_wait();
    }
    
    exit(); // trigger reset
}

void irq()
{
    // not used
}

// moved to music section above
//void nmi_music()
//{
//}

#endif // !NSF

// NSF Main
// ========

void nsf_init()
{
    prng_seed = 1;
    setup_music();
}

void nsf_play()
{
    nmi_sound();
    tick_music();
}

// end of file